"""
  Module with methods for calculating creep-fatigue damage given 
  completely-solved tube results and damage material properties
"""

import numpy as np
import scipy.optimize as opt

import multiprocess

class DamageCalculator:
  """
    Parent class for all damage calculators, handling common iteration
    and scaling options
  """
  def __init__(self):
    pass

  def single_cycles(self, tube, material, receiver):
    """
      Calculate damage for a single tube

      Parameters:
        tube:       fully-populated tube object
        material:   damage material
        receiver:   receiver object (for metadata)
    """ 
    raise NotImplementedError("Superclass not implemented")

  def determine_life(self, receiver, material, nthreads = 1, 
      decorator = lambda x, n: x):
    """
      Determine the life of the receiver by calculating individual 
      material point damage and finding the minimum of all points.

      Parameters:
        receiver        fully-solved receiver object
        material        material model ot use

      Additional Parameters:
        nthreads        number of threads
        decorator       progress bar
    """
    # pylint: disable=no-member
    with multiprocess.Pool(nthreads) as p:
      Ns = list(decorator(
        p.imap(lambda x: self.single_cycles(x, material, receiver), receiver.tubes),
        receiver.ntubes))
    N = min(Ns)

    # Translate to days
    return N * receiver.days

class TimeFractionInteractionDamage(DamageCalculator):
  """
    Calculate life using the ASME time-fraction type approach
  """
  def single_cycles(self, tube, material, receiver):
    """
      Calculate the single-tube number of repetitions to failure

      Parameters:
        tube        single tube with full results
        material    damage material model
        receiver    receiver, for metadata
    """
    # Material point creep damage
    Dc = self.creep_damage(tube, material)

    # Material point fatigue damage
    Df = self.fatigue_damage(tube, material, receiver)

    # This is going to be expensive, but I don't see much way around it
    return min(self.calculate_max_cycles(c, f, material) for c,f in 
        zip(Dc.flatten(), Df.flatten()))
  
  def calculate_max_cycles(self, Dc, Df, material, rep_min = 1, rep_max = 1e6):
    """
      Actually calculate the maximum number of repetitions for a single point

      Parameters:
        Dc          creep damage per simulated cycle
        Df          fatigue damage per simulated cycle
        material    damaged material properties
    """
    if np.isclose(Dc, 0) and np.isclose(Df, 0):
      return np.inf

    if not material.inside_envelope("cfinteraction", rep_min * Df, rep_min*Dc):
      return 0

    if material.inside_envelope("cfinteraction", rep_max*Df, rep_max * Dc):
      return np.inf

    return opt.brentq(lambda N: material.inside_envelope("cfinteraction", N* Df, N*Dc) - 0.5,
        rep_min, rep_max)

  def creep_damage(self, tube, material):
    """
      Calculate creep damage at each material point

      Parameters:
        tube        single tube with full results
        material    damage material model
    """
    # For now just use the von Mises effective stress
    vm = np.sqrt((
        (tube.quadrature_results['stress_xx'] - tube.quadrature_results['stress_yy'])**2.0 + 
        (tube.quadrature_results['stress_yy'] - tube.quadrature_results['stress_zz'])**2.0 + 
        (tube.quadrature_results['stress_zz'] - tube.quadrature_results['stress_xx'])**2.0 + 
        3.0 * (tube.quadrature_results['stress_xy']**2.0 + 
          tube.quadrature_results['stress_yz']**2.0 + 
          tube.quadrature_results['stress_xz']**2.0))/2.0)

    tR = material.time_to_rupture("averageRupture", tube.quadrature_results['temperature'], vm)
    dts = np.diff(tube.times)
    dmg = np.sum(dts[:,np.newaxis,np.newaxis]/tR[1:], axis = 0)

    return dmg

  def fatigue_damage(self, tube, material, receiver):
    """
      Calculate fatigue damage at each material point

      Parameters:
        tube        single tube with full results
        material    damage material model
        receiver    receiver, for metadata
    """
    # Identify cycle boundaries
    tm = np.mod(tube.times, receiver.period)
    inds = list(np.where(tm == 0)[0])
    if len(inds) != (receiver.days + 1):
      raise ValueError("Tube times not compatible with the receiver"
          " number of days and cycle period!")

    # Run through each cycle and ID max strain range and fatigue damage
    strain_names = ['mechanical_strain_xx', 'mechanical_strain_yy', 'mechanical_strain_zz',
        'mechanical_strain_yz', 'mechanical_strain_xz', 'mechanical_strain_xy']
    strain_factors = [1.0,1.0,1.0,np.sqrt(2), np.sqrt(2), np.sqrt(2)]
    
    return sum(self.cycle_fatigue(np.array([ef*tube.quadrature_results[en][
      inds[i]:inds[i+1]] for 
      en,ef in zip(strain_names, strain_factors)]), 
      tube.quadrature_results['temperature'][inds[i]:inds[i+1]], material)
      for i in range(receiver.days))
  
  def cycle_fatigue(self, strains, temperatures, material, nu = 0.5):
    """
      Calculate fatigue damage for a single cycle

      Parameters:
        strains         single cycle strains
        temperatures    single cycle temperatures
        material        damage model

      Additional parameters:
        nu              effective Poisson's ratio to use
    """
    pt_temps = np.max(temperatures, axis = 0)

    pt_eranges = np.zeros(pt_temps.shape)
    nt = strains.shape[0]
    for i in range(nt):
      for j in range(nt):
        de = strains[:,j] - strains[:,i]
        eq = np.sqrt(2) / (2*(1+nu)) * np.sqrt(
            (de[0] - de[1])**2 + (de[1]-de[2])**2 + (de[2]-de[0])**2.0
            + 3.0/2.0 * (de[3]**2.0 + de[4]**2.0 + de[5]**6.0)
            )
        pt_eranges = np.maximum(pt_eranges, eq)
    
    dmg = np.zeros(pt_eranges.shape)
    # pylint: disable=not-an-iterable
    for ind in np.ndindex(*dmg.shape):
      dmg[ind] = 1.0 / material.cycles_to_fail("nominalFatigue", pt_temps[ind], pt_eranges[ind])

    return dmg



import unittest

from srlife import library

fluids = ['salt']
alloys = ['316H','800H','A617','740H','A282','A230']

class TestMaterials(unittest.TestCase):
  def test_load_materials(self):
    for alloy in alloys:
      thermal, deformation, damage = library.load_material(alloy, "base","base","base")
      deformation.get_neml_model()
      thermal, deformation, damage = library.load_material(alloy, "base","elastic_creep","base")
      deformation.get_neml_model()
      thermal, deformation, damage = library.load_material(alloy, "base","elastic_model","base")
      deformation.get_neml_model()

    def test_load_fluids(self):
      for fluid in fluids:
        library.load_fluid(fluid,"base")



import unittest

import os.path

# Shut up matplotlib
import logging
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING) 

import numpy as np

import sys
sys.path.append("test/solvers/tube")
sys.path.append("../../..")
from srlife import structural, receiver

from neml import elasticity, models, parse

from run_structural_tube_verification import cases, do_complete_comparison

def differentiate(fn, x0, eps = 1.5e-6):
  if np.isscalar(x0):
    sx = (1,)
  else:
    sx = x0.shape

  f0 = fn(x0)

  if np.isscalar(f0):
    sf = (1,)
  else:
    sf = f0.shape

  tshape = sf + sx
  rshape = sx + sf

  D = np.zeros(rshape)

  for index in np.ndindex(sx):
    if np.isscalar(x0):
      diff = np.abs(x0 * eps)
    else:
      diff = np.abs(x0[index] * eps)
    if diff < eps:
      diff = eps
    
    if np.isscalar(x0):
      xp = diff
    else:
      xp = np.zeros(sx)
      xp[index] = diff

    fp = fn(xp+x0)

    D[index] = (fp - f0) / diff
  
  Df = np.zeros(tshape)
  # Reverse
  for ind1 in np.ndindex(sx):
    for ind2 in np.ndindex(sf):
      Df[ind2+ind1] = D[ind1+ind2]
  
  if np.isscalar(x0):
    return Df[0][0]
  else:
    return Df

class TestSimpleCases(unittest.TestCase):
  def setUp(self):
    self.atol = 1.0e-3
    self.rtol = 1.0e-3

    self.problems = cases
    self.solver = structural.PythonTubeSolver()

  def _check_case(self, d, case):
    tube = case.run_comparison(d, self.solver)
    a, r = case.evaluate_comparison(tube)

    self.assertTrue(a < self.atol or r < self.rtol)
  
  def test_all(self):
    for d in range(1,4):
      for c in self.problems:
        self._check_case(d, c)

class TestMOOSEComparison(unittest.TestCase):
  def setUp(self):
    self.solver = structural.PythonTubeSolver()
    self.atol = 1.0e-3

  def test_1D(self):
    a, r = do_complete_comparison(1, self.solver)
    self.assertTrue(a < self.atol)

  def test_2D(self):
    a, r = do_complete_comparison(2, self.solver)
    self.assertTrue(a < self.atol)
  
  # Too heavy right now
  #def test_3D(self):
  #  a, r = do_complete_comparison(3, self.solver)
  #  self.assertTrue(a < self.atol)

class TestAxialStiffnessExact(unittest.TestCase):
  def setUp(self):
    self.ro = 5
    self.ri = 4.5
    self.h = 2.5

    self.nr = 10
    self.nt = 20
    self.nz = 5

    self.E = 150000.0
    self.nu = 0.35

    self.tube = receiver.Tube(self.ro, self.ro - self.ri, self.h, self.nr, self.nt, self.nz)
    self.times = np.array([0,1])
    self.tube.set_times(self.times)

    self.tube.set_pressure_bc(receiver.PressureBC(self.times, self.times * 0))

    self.emodel = elasticity.IsotropicLinearElasticModel(self.E, "youngs", self.nu, "poissons")
    self.mat = models.SmallStrainElasticity(self.emodel)

    self.d = 0.25

    self.force_exact = np.pi * (self.ro**2.0 - self.ri**2.0) * self.E * self.d / self.h
    self.stiffness_exact = np.pi * (self.ro**2.0 - self.ri**2.0) * self.E / self.h

    self.solver = structural.PythonTubeSolver(verbose = False) 

  def _solve(self, d):
    self.solver.setup_tube(self.tube)
    state_n = self.solver.init_state(self.tube, self.mat)

    return self.solver.solve(self.tube, 1, state_n, d)

  def test_1D(self):
    self.tube.make_1D(self.h/ 2, 0)

    state = self._solve(self.d)

    self.assertAlmostEqual(self.force_exact, state.force, places = 3)
    self.assertAlmostEqual(self.stiffness_exact, state.stiffness, places = 3)

  def test_2D(self):
    self.tube.make_2D(self.h/ 2)

    state = self._solve(self.d)

    cf = np.sum(state.basis.dx) / (np.pi * (self.ro**2.0 - self.ri**2.0))

    self.assertAlmostEqual(self.force_exact, state.force / cf, places = 3)
    self.assertAlmostEqual(self.stiffness_exact, state.stiffness / cf, places = 3)

  def test_3D(self):
    state = self._solve(self.d)
    
    cf = np.sum(state.basis.dx) / self.h / (np.pi * (self.ro**2.0 - self.ri**2.0))    

    self.assertAlmostEqual(self.force_exact, state.force / cf, places = 3)
    self.assertAlmostEqual(self.stiffness_exact, state.stiffness / cf, places = 3)

class TestAxialStiffnessNumerical(unittest.TestCase):
  def setUp(self):
    self.ro = 5
    self.ri = 4.5
    self.h = 2.5

    self.nr = 10
    self.nt = 20
    self.nz = 5

    self.E = 150000.0
    self.nu = 0.35

    self.tube = receiver.Tube(self.ro, self.ro - self.ri, self.h, self.nr, self.nt, self.nz)
    self.times = np.array([0,1])
    self.tube.set_times(self.times)

    self.tube.set_pressure_bc(receiver.PressureBC(self.times, self.times / 50.0))

    self.mat = parse.parse_xml(os.path.join(os.path.dirname(__file__),
      'moose-verification', 'model.xml'), 'creeping')

    self.d = 0.25

    self.solver = structural.PythonTubeSolver(verbose = False) 

  def _solve(self, d):
    self.solver.setup_tube(self.tube)
    state_n = self.solver.init_state(self.tube, self.mat)

    return self.solver.solve(self.tube, 1, state_n, d)

  def test_1D(self):
    self.tube.make_1D(self.h/ 2, 0)

    exact = self._solve(self.d).stiffness
    numerical = differentiate(lambda d: self._solve(d).force, self.d)

    self.assertTrue(np.isclose(exact, numerical, rtol = 1e-4))

  def test_2D(self):
    self.tube.make_2D(self.h/ 2)

    exact = self._solve(self.d).stiffness
    numerical = differentiate(lambda d: self._solve(d).force, self.d)

    self.assertTrue(np.isclose(exact, numerical, rtol = 1e-4))

# Too heavy
  def test_3D(self):
    exact = self._solve(self.d).stiffness
    numerical = differentiate(lambda d: self._solve(d).force, self.d)

    self.assertTrue(np.isclose(exact, numerical, rtol = 1e-4))



# Weirdness with pylint and vtk (some dynamic import thing going on there)
# pylint: disable=no-member
"""
  Module for various ways to write receiver data structures to disk in a human-readable way
"""
import vtk

import numpy as np

class VTKWriter:
  """
    Write a single tube file to VTK
  """
  def __init__(self, tube, fname):
    """
      Parameters:
        tube        tube object
        fname       base filename to use
    """
    self.tube = tube
    self.fname = fname

  def write(self):
    """
      Actually write the tube object to a vtk file using a
      VTKStructuredGrid
    """
    grid = vtk.vtkUnstructuredGrid()

    R, T, Z = self.tube.mesh
    points = vtk.vtkPoints()

    X = R * np.cos(T)
    Y = R * np.sin(T)

    for x,y,z in zip(X.flatten(), Y.flatten(), Z.flatten()):
      points.InsertNextPoint(x,y,z)

    grid.SetPoints(points)
    
    self._set_grid(grid)
    
    writer = vtk.vtkXMLUnstructuredGridWriter()
    writer.SetFileName(self.fname + ".vtu")
    writer.SetInputData(grid)  
    writer.SetNumberOfTimeSteps(self.tube.ntime)

    writer.Start()
    for i in range(self.tube.ntime):
      self._dump_point_data(grid, i)
      self._dump_element_data(grid, i)
      writer.WriteNextTime(self.tube.times[i])
    writer.Stop()
  
  def _dump_point_data(self, grid, i):
    """
      Dump all our point data to the grid

      Parameters:
        grid        vtk unstructured grid
        i           time step
    """
    for field, data in self.tube.results.items():
      pdata = vtk.vtkFloatArray()
      pdata.SetNumberOfComponents(1)
      pdata.SetName(field)
      for d in data[i].flatten():
        pdata.InsertNextValue(d)
      grid.GetPointData().AddArray(pdata)

  def _dump_element_data(self, grid, i):
    for field, data in self.tube.quadrature_results.items():
      pdata = vtk.vtkFloatArray()
      pdata.SetNumberOfComponents(1)
      pdata.SetName(field)
      avg = np.mean(data[i], axis = 1)
      for d in avg:
        pdata.InsertNextValue(d)
      grid.GetCellData().AddArray(pdata)
    
  def _set_grid(self, grid):
    """
      Setup the VTK unstructured grid for the mesh

      Parameters:
        grid        vtk unstructured grid
    """
    if self.tube.ndim == 1:
      self._set_grid_1d(grid)
    elif self.tube.ndim == 2:
      self._set_grid_2d(grid)
    else:
      self._set_grid_3d(grid)

  def _set_grid_1d(self, grid):
    """
      Setup the VTK unstructured grid for a 1d mesh

      Parameters:
        grid        vtk unstructured grid
    """
    flat = lambda i: i
    for i in range(self.tube.nr-1):
      cell = vtk.vtkLine()
      for nk, a in enumerate((0,1)):
        cell.GetPointIds().SetId(nk, flat(i+a))
      grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())

  def _set_grid_2d(self, grid):
    """
      Setup the VTK unstructured grid for a 1d mesh

      Parameters:
        grid        vtk unstructured grid
    """
    flat = lambda i, j: i * self.tube.nt + (j % self.tube.nt)
    for i in range(self.tube.nr-1):
      for j in range(self.tube.nt):
        cell = vtk.vtkQuad()
        for nk, (a,b) in enumerate(((0,0),(0,1),(1,1),(1,0))):
          cell.GetPointIds().SetId(nk, flat(i+a,j+b))
        grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())

  def _set_grid_3d(self, grid):
    """
      Setup the VTK unstructured grid for a 1d mesh

      Parameters:
        grid        vtk unstructured grid
    """
    flat = lambda i, j, k: i * self.tube.nt * self.tube.nz + (j % self.tube.nt) * self.tube.nz + k
    for i in range(self.tube.nr-1):
      for j in range(self.tube.nt):
        for k in range(self.tube.nz-1):
          cell = vtk.vtkHexahedron()
          for nk, (a,b,c) in enumerate(
              ((0,0,0),(1,0,0),(1,1,0),(0,1,0),(0,0,1),(1,0,1),(1,1,1),(0,1,1))):
            cell.GetPointIds().SetId(nk,flat(i+a,j+b,k+c))
          grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())



"""
  Nonlinear solvers and other accessories
"""

import numpy as np
import numpy.linalg as la

def newton(RJ, x0, rel_tol = 1.0-6, abs_tol = 1.0e-8, miters = 20,
    linear_solver = la.solve, verbose = True, return_extra = False):
  """
    Simple newton-raphson solver

    Parameters:
      RJ            function that gives the residual and jacobian values
      x0            initial guess

    Additional parameters:
      reL_tol       relative convergence tolerance
      abs_tol       absolute convergence tolerance
      miters        maximum number of iterations
      linear_solver function that solves the linear system A x = b
      verbose       if true, print debug info,
      return_extra  if true also return the final residual vector and
                    Jacobian matrix
  """
  x = np.copy(x0)

  R, J = RJ(x)

  nR = la.norm(R)
  nR0 = nR

  if verbose:
    print("Iter\tnR\t\tnR/nR0")
    print("%i\t%3.2e" % (0, nR0))

  for i in range(miters):
    if nR < abs_tol or nR/nR0 < rel_tol:
      break
    x -= linear_solver(J, R)
    R, J = RJ(x)    
    nR = la.norm(R)
    if verbose:
      print("%i\t%3.2e\t%3.2e" % (i+1, nR, nR/nR0))
  else:
    raise RuntimeError("Nonlinear solver did not converge!")

  if verbose:
    print("")

  if return_extra:
    return x, R, J
  else:
    return x



import unittest

import numpy as np

from srlife import spring

class TestSimple(unittest.TestCase):
  """
    Simple test cases with elastic springs
  """
  def test_single(self):
    network = spring.SpringNetwork()
    network.add_node(0)
    network.add_node(1)
    network.add_edge(0,1, object = spring.LinearSpring(100))
    
    network.set_times([0,1])

    network.validate_setup()

    network.displacement_bc(0, lambda t: 0.0)
    network.force_bc(1, lambda t: 100.0*t)

    network.solve_all()

    self.assertTrue(np.allclose(network.displacements,
      [0,1]))

  def test_system(self):
    k1 = 100.0
    k2 = 200.0
    k3 = 150.0
    f = 100.0

    network = spring.SpringNetwork()
    network.add_node(0)
    network.add_node(1)
    network.add_node(2)
    network.add_edge(0,1, object = spring.LinearSpring(k1))
    network.add_edge(1,2, object = spring.LinearSpring(k2))
    network.add_edge(1,2, object = spring.LinearSpring(k3))
    
    network.set_times([0,1])

    network.validate_setup()

    network.displacement_bc(2, lambda t: 0.0)
    network.force_bc(0, lambda t: t*f)

    network.solve_all()

    kp = k2 + k3
    dxp = f / kp
    dx1 = f / k1
    
    exact = np.array([dx1+dxp,dxp,0])

    self.assertTrue(np.allclose(network.displacements,exact))

  def test_system_thrads(self):
    k1 = 100.0
    k2 = 200.0
    k3 = 150.0
    f = 100.0

    network = spring.SpringNetwork()
    network.add_node(0)
    network.add_node(1)
    network.add_node(2)
    network.add_edge(0,1, object = spring.LinearSpring(k1))
    network.add_edge(1,2, object = spring.LinearSpring(k2))
    network.add_edge(1,2, object = spring.LinearSpring(k3))
    
    network.set_times([0,1])

    network.validate_setup()

    network.displacement_bc(2, lambda t: 0.0)
    network.force_bc(0, lambda t: t*f)

    network.solve_all(nthreads=2)

    kp = k2 + k3
    dxp = f / kp
    dx1 = f / k1
    
    exact = np.array([dx1+dxp,dxp,0])

    self.assertTrue(np.allclose(network.displacements,exact))

  def test_remove_rigid(self):
    network = spring.SpringNetwork()
    k = 100.0
    
    for i in range(6):
      network.add_node(i)

    network.add_edge(0,1, object = "rigid")
    network.add_edge(1,3, object = spring.LinearSpring(k))
    network.add_edge(1,3, object = spring.LinearSpring(k))
    network.add_edge(2,4, object = "rigid")
    network.add_edge(3,5, object = spring.LinearSpring(k))
    network.add_edge(4,5, object = spring.LinearSpring(k))
    
    network.displacement_bc(4, lambda x: 0)

    network.set_times([0,1])
    network.validate_setup()
    
    network.remove_rigid()
    network.validate_solve()
    
    self.assertEqual(len(network.nodes), 4)
    self.assertEqual(len(network.edges), 4)

  def test_disconnect(self):
    network = spring.SpringNetwork()
    k = 100.0
    
    for i in range(8):
      network.add_node(i)

    network.add_edge(0, 1, object = spring.LinearSpring(k))
    network.add_edge(0, 2, object = spring.LinearSpring(k))

    network.add_edge(1, 3, object = spring.LinearSpring(k))
    network.add_edge(1, 3, object = spring.LinearSpring(k))
    network.add_edge(2, 4, object = "disconnect")

    network.add_edge(3, 5, object = spring.LinearSpring(k))
    network.add_edge(4, 6, object = spring.LinearSpring(k))
    network.add_edge(4,7, object = spring.LinearSpring(k))

    
    network.displacement_bc(0, lambda x: 0)
    network.displacement_bc(4, lambda x: 0)

    network.set_times([0,1])
    network.validate_setup()

    subs = network.reduce_graph()

    self.assertEqual(len(subs), 2)

    for sub in subs:
      sub.validate_solve()



#!/usr/bin/env python3

import os.path

import numpy as np
import numpy.linalg as la
import scipy.io as sio

import matplotlib.pyplot as plt
from neml import models, elasticity, parse

import sys
sys.path.append('../../..')
from srlife import receiver, structural

class TestCase:
  def __init__(self, name, T, analytic, ri = 0.9, ro = 1.0, h = 10.0,
      alpha = 1.0e-5, E = 100000.0, nu = 0.3, p = 1.0):
    self.name = name
    self.Tfn = T
    self.afn = analytic

    self.ri = ri
    self.ro = ro
    self.h = h

    self.alpha = alpha
    self.E = E
    self.nu = nu

    self.p = p

  def T(self, r):
    return self.Tfn(r, self.ri, self.ro)

  def exact(self, r):
    return self.afn(r, self.p, self.ri, self.ro, self.E, self.nu, self.alpha)

  def make_mat(self):
    emodel = elasticity.IsotropicLinearElasticModel(self.E, "youngs", 
        self.nu, "poissons")
    return models.SmallStrainElasticity(emodel, alpha = self.alpha)

  def make_tube(self, dim, nr = 15, nt = 30, nz = 5):
    tube = receiver.Tube(self.ro, self.ro - self.ri, self.h, nr, nt, nz)

    if dim == 1:
      tube.make_1D(self.h/2, 0)
    elif dim == 2:
      tube.make_2D(self.h/2)

    times = np.array([0,1])
    tube.set_times(times)

    R, _, _ = tube.mesh
    Ts = np.zeros((2,) + R.shape[:dim])
    Ts[1] = self.T(R)

    tube.add_results("temperature", Ts)

    if self.p != 0:
      tube.set_pressure_bc(receiver.PressureBC(times, times * self.p))

    return tube

  def run_comparison(self, dim, solver, axial_strain = 0, nr = 10, 
      nt = 20, nz = 10):
    mat = self.make_mat()
    tube = self.make_tube(dim, nr, nt, nz)
    
    solver.setup_tube(tube)
    state_n = solver.init_state(tube, mat)

    state_np1 = solver.solve(tube, 1, state_n, axial_strain)

    solver.dump_state(tube, 1, state_np1)

    return tube

  def get_comparison(self, tube):
    if tube.ndim == 3:
      z = tube.nz // 2
      x_avg = np.mean(tube.results['disp_x'])
      u = tube.results['disp_x'][1,:,0,z] - 2*x_avg
      r = tube.mesh[0][:,0,z]
    elif tube.ndim == 2:
      # The displacements tend to drift, need to recenter
      x_avg = np.mean(tube.results['disp_x'])
      u = tube.results['disp_x'][1,:,0] - 2*x_avg
      r = tube.mesh[0][:,0,0]
    else:
      u = tube.results['disp_x'][1]
      r = tube.mesh[0][:,0,0]

    return u, r

  def plot_comparison(self, tube):
    u, r = self.get_comparison(tube)

    plt.figure()
    plt.plot(r, u, 'k-')
    plt.plot(r, self.exact(r), 'k--')
    plt.xlabel("Radial position")
    plt.ylabel("Radial displacement")
    plt.title(self.name + ": " + "%iD" % tube.ndim)
    plt.show()

  def evaluate_comparison(self, tube):
    u, r = self.get_comparison(tube)
    
    err = np.abs(u - self.exact(r))
    rel = err / np.abs(self.exact(r))

    return np.max(err), np.max(rel)


def exact1(r, p, ri, ro, E, v, a):
  A = p / (1.0/ro**2.0 - 1.0/ri**2.0)
  C = -A/(2*ro**2.0)

  res = (1.0+v)/E * (-A/r + 2.0*(1-2.0*v) * C* r)

  return res

cases = [
    TestCase("Inner pressure", lambda T, ri, ro: 0.0, exact1,
      p = 100, ri = 8, ro = 10.0)
    ]


# Moose comparison stuff
# MOOSE test stuff
ramp = lambda x: np.piecewise(x, 
    [x < 1, x>=1],
    [lambda xx: xx, lambda xx: 1.0+xx*0.0])
unit_temperature = lambda x, y, z: (np.sqrt(x**2.0+y**2.0)-9)/(10-9) * np.cos(np.arctan2(y,x))*(z/10.0 + 1)
temperature = lambda t, x, y, z: np.array([100*ramp(ti)*unit_temperature(x,y,z) for ti in t])
pressure = lambda t: 1.0 * ramp(t)

times = np.array([0, 0.5, 1, 101, 201, 301, 401, 501, 601, 701, 801, 901, 1001])
mat = parse.parse_xml(os.path.join(os.path.dirname(__file__),
  'moose-verification', 'model.xml'), 'creeping')

ri = 9.0
ro = 10.0
h = 10.0
nr = 11
nt = 20
nz = 6

moose_base = os.path.join(os.path.dirname(__file__), 'moose-verification')
moose_ver = [os.path.join(moose_base, f) for f in ['1d_out.e', '2d_out.e', '3d_out.e']]

def run_reference_simulation(d, solver):
  tube = receiver.Tube(ro, ro - ri, h, nr, nt, nz)
  tube.set_times(times)

  if d == 1:
    tube.make_1D(0, 0)
  elif d == 2:
    tube.make_2D(0)

  R, T, Z = tube.mesh
  X = R * np.cos(T)
  Y = R * np.sin(T)
  
  Ts = temperature(times, X, Y, Z).reshape((len(times),) + tube.dim[:tube.ndim])

  tube.add_results("temperature", Ts)
  tube.set_pressure_bc(receiver.PressureBC(times, pressure(times)))

  solver.setup_tube(tube)

  state_n = solver.init_state(tube, mat)

  for i in range(1,len(tube.times)):
    state_np1 = solver.solve(tube, i, state_n, 0.0)
    solver.dump_state(tube, i, state_np1)
    state_n = state_np1
  
  return tube

def load_displacements_exodus(efilename, d):
  nf = sio.netcdf_file(efilename)
  cnames = ['coordx', 'coordy', 'coordz']
  coords = np.array([
    np.copy(nf.variables[c][:]) for c in cnames[:d]])

  names = [b''.join(v).decode('utf-8') for v in nf.variables['name_nod_var']]
  disps = np.array([np.copy(nf.variables['vals_nod_var%i' % (names.index(nm)+1)][:]) for 
      nm in ['disp_x', 'disp_y', 'disp_z'][:d]])

  return coords, disps

def compare_nodal_field(c1, r1, c2, r2, d, dec = 6, gtol = 1.0e-8):
  c1 = np.round(c1, decimals=dec)
  c2 = np.round(c2, decimals=dec)
  for i in range(d-1,-1,-1):
    one_args = c1[i,:].argsort(kind='mergesort')
    c1 = c1[:,one_args]

    two_args = c2[i,:].argsort(kind='mergesort')
    c2 = c2[:,two_args]

    for j in range(d):
      for k in range(r1.shape[1]):
        r1[j,k] = r1[j,k,one_args]
        r2[j,k] = r2[j,k,two_args]

  # Sanity check
  if la.norm(c1-c2) > gtol:
    raise ValueError("Point clouds aren't equivalent...")

  error = np.abs(r1 - r2)
  aerror = np.max(error)

  keep = np.abs(r1) > gtol
  rerror = np.max(error[keep] / np.abs(r1[keep]))

  return aerror, rerror

def do_complete_comparison(d, solver):
  tube = run_reference_simulation(d, solver)

  R, T, Z = tube.mesh
  X = R * np.cos(T)
  Y = R * np.sin(T) 
  cs = [X,Y,Z]
  coords = np.array([c.flatten() for c in cs[:d]])
  dres = ['disp_x', 'disp_y', 'disp_z']
  disps = np.array([tube.results[d].reshape((len(tube.times),-1)) for d in dres[:d]])

  mc, md = load_displacements_exodus(moose_ver[d-1], d)

  return compare_nodal_field(mc, md, coords, disps, d)


if __name__ == "__main__":
  solver = structural.PythonTubeSolver(verbose = False)

  print("Analytical comparison")
  print("=====================")
  print("")
  for d in range(1,4):
    for case in cases:
      tube = case.run_comparison(d, solver)

      a, r = case.evaluate_comparison(tube)
      print(case.name + ": " "%iD" % d)
      print("Max absolute error: %e" % a)
      print("Max relative error: %e" % r)
      print("")

      case.plot_comparison(tube)
  
  print("MOOSE comparison")
  print("================")
  print("")
  for d in range(1,4):
    a, r = do_complete_comparison(d,solver)
    print("%iD" % d)
    print("Max absolute error: %e" % a)
    print("")



#!/usr/bin/env python3

import sys
sys.path.append('../../..')
from srlife import materials, thermal

from thermalsol import ManufacturedSolution

import numpy as np

import matplotlib.pyplot as plt

problems = [
    ManufacturedSolution("1D: no spatial", 1, 
      lambda t, r: t, 
      lambda t, k, alpha, r: k/alpha * (r*0.0+1)),
    ManufacturedSolution("1D: spatial and time", 1, 
      lambda t, r: np.sin(t)*np.log(r), 
      lambda t, k, alpha, r: k/alpha * np.log(r) * np.cos(t)),
    ManufacturedSolution("1D: just spatial", 1,
      lambda t, r: np.sin(r),
      lambda t, k, alpha, r: k * np.sin(r) - k/r*np.cos(r)),
    ManufacturedSolution("2D: just time", 2,
      lambda t, r, th: t,
      lambda t, k, alpha, r, th: k/alpha * (r*0.0+1)),
    ManufacturedSolution("2D: just r", 2,
      lambda t, r, th: np.sin(r),
      lambda t, k, alpha, r, th: k * np.sin(r) - k/r * np.cos(r)),
    ManufacturedSolution("2D: just theta", 2,
      lambda t, r, th: np.cos(th),
      lambda t, k, alpha, r, th: k * np.cos(th) / r**2.0),
    ManufacturedSolution("2D: theta and r", 2,
      lambda t, r, th: np.cos(th) / r,
      lambda t, k, alpha, r, th: -k*np.cos(th) / r**3.0),
    ManufacturedSolution("2D: all three", 2,
      lambda t, r, th: np.log(r) * np.sin(th) / (t+1),
      lambda t, k, alpha, r, th: k*np.log(r)*np.sin(th)/((t+1)*r**2.0) - k/alpha * np.log(r) * np.sin(th) / (t+1)**2.0),
    ManufacturedSolution("3D: just t", 3,
      lambda t, r, th, z: t,
      lambda t, k, alpha, r, th, z: k/alpha * (r*0.0+1)),
    ManufacturedSolution("3D: just r", 3,
      lambda t, r, th, z: np.sin(r),
      lambda t, k, alpha, r, th, z: k * np.sin(r) - k/r * np.cos(r)),
    ManufacturedSolution("3D: just theta", 3,
      lambda t, r, th, z: np.cos(th),
      lambda t, k, alpha, r, th, z: k * np.cos(th) / r**2.0),
    ManufacturedSolution("3D: just z", 3,
      lambda t, r, th, z: np.sin(z),
      lambda t, k, alpha, r, th, z: k * np.sin(z)),
    ManufacturedSolution("3D: all spatial", 3,
      lambda t, r, th, z: z**2.0*np.cos(th)/r,
      lambda t, k, alpha, r, th, z: -k*np.cos(th)/r*((z/r)**2.0+2)),
    ManufacturedSolution("3D: everything", 3,
      lambda t, r, th, z: np.log(r)*np.sin(th)*np.cos(z)/(t+1.0),
      lambda t, k, alpha, r, th, z: k*np.log(r) * np.sin(th) * np.cos(z) / (t+1) * (1.0 + 1/r**2.0 - 1.0/(alpha*(t+1)))),
    ]


def run_with(solver, material, fluid):
  """
    Run the standard problems with the provided solver and material
  """
  for problem in problems:
    res = problem.solve(solver, material, fluid)
    problem.plot_comparison(res)
    plt.show()

if __name__ == "__main__":
  run_with(thermal.FiniteDifferenceImplicitThermalSolver(), 
      materials.ConstantThermalMaterial("Test", 10.0, 5.0),
      materials.ConstantFluidMaterial({"Test": 7.5}))



import unittest

import numpy as np
import numpy.random as ra

from srlife import helpers

class TestRandom(unittest.TestCase):
  def test_ms2ts(self):
    C = ra.random((6,6))
    C1 = helpers.ms2ts(C)
    C2 = helpers.ms2ts_faster(C)

    self.assertTrue(np.allclose(C1,C2))

  def test_usym(self):
    M = ra.random((6,))
    M1 = helpers.usym(M)
    M2 = helpers.usym_faster(M)

    self.assertTrue(np.allclose(M1,M2))

  def test_sym(self):
    T = ra.random((3,3))
    T = 0.5*(T+T.T)

    M1 = helpers.sym(T)
    M2 = helpers.sym_faster(T)

    self.assertTrue(np.allclose(M1,M2))
